%{					/* -*-c-*- */
 /**********************************************************************
 @Lex-file{
    author              = "Nelson H. F. Beebe",
    version             = "0.11",
    date                = "25 April 1996",
    time                = "18:05:46 MDT",
    filename            = "htmlpty.l",
    address             = "Center for Scientific Computing
                           Department of Mathematics
                           University of Utah
                           Salt Lake City, UT 84112
                           USA",
    telephone           = "+1 801 581 5254",
    FAX                 = "+1 801 581 4148",
    checksum            = "53814 2522 6407 54722",
    email               = "beebe@math.utah.edu (Internet)",
    codetable           = "ISO/ASCII",
    keywords            = "HTML, prettyprinter, SGML, WWW, World-Wide Web",
    supported           = "yes",
    docstring           = "This is a simple HTML file prettyprinter.

                           Usage:
                               html-pretty [-a] [-c] [-f filename] [-h]
                                           [-i nnn] [-n] [-v] [-w width]
                                           [file(s)]

			   The options are -a (show author), -c (show
			   copyright), -f (set output filename in
			   comment banner), -i (set indentation size),
			   -n (no comment banner), -v (show version),
			   and -w (set linewidth).  Letter case in
			   option names is NOT significant.  On IBM PC
			   DOS, the leading hyphen in option names may
			   be replaced by a slash; both hyphen and slash
			   are accepted.

                           The idea is that each <TAG> ... </TAG> pair
                           should contain material that is indented
                           according to the nesting depth of the tags.
                           However, certain common tags, such as those
                           to change fonts, are excluded from this
                           indenting.

                           As with any simple-minded prettyprinter,
                           especially ones for natural language text,
                           which rarely conforms to a strict grammar,
                           you should check the output of the
                           prettyprinter before replacing the input by
                           it.  In particular, if your input file
                           follows the standards imposed by this
                           prettyprinter, a file difference listing may
                           be reassuring.  Otherwise, try viewing the
                           output with one or more World-Wide Web client
                           programs (arena, hotjava, lynx, netscape,
                           xmosaic, ...).  If you have an SGML parser or
                           translator, you can use that for checking
                           too.

                           The checksum field above contains a CRC-16
                           checksum as the first value, followed by the
                           equivalent of the standard UNIX wc (word
                           count) utility output of lines, words, and
                           characters.  This is produced by Robert
                           Solovay's checksum utility.",
 }
 **********************************************************************/

	/* htmlpty -- prettyprint World-Wide Web HTML files */

#define VERSION		"0.11"		/* these MUST match version */
#define DATE		"[25-Apr-1996]"	/* and date above */
#define PROGRAM		"html-pretty"

#if !defined(__DATE__)
#define __DATE__ ""
#endif /* __DATE__ */

#if !defined(__TIME__)
#define __TIME__ ""
#endif /* __TIME__ */

#if !defined(USER)
#define USER	""
#endif

#if !defined(HOSTNAME)
#define HOSTNAME ""
#endif

#include <stdio.h>
#include <string.h>			/* for strlen() declaration */
#include <ctype.h>

#if defined(unix) || defined(_unix) || defined(__unix) || defined(__unix__) \
	|| defined(_IBMR2)
#define UNIX
#include <sys/types.h>
#include <pwd.h>
#if defined(__NeXT__)
#undef HAVE_UNAME
#if defined(__GNUG__)
#include <libc.h>
#else
/***********************************************************************
This ugliness of explicit declarations arises because the installation
of gcc 2.6.3 (and earlier) on the NeXT system searches vendor-provided
header files before gcc ones; the result is function prototype
conflicts.  They can be avoided if the code is compiled with an explicit
include-file path, such as
	-I/usr/local/lib/gcc-lib/m68k-next-nextstep3/2.6.3/include/bsd
but that is an ugly solution too, so instead we just provide copies of
the declarations from the gcc libc.h that we actually need.
***********************************************************************/
extern char	*getlogin(void);
extern uid_t    getuid(void);
extern int      gethostname(char *, int);
#endif
#else
#define HAVE_UNAME
#include <unistd.h>
#endif
#else
#undef UNIX
#endif

#if defined(HAVE_UNAME)
#include <sys/utsname.h>
#endif

#include <time.h>

#define COMMENT_PREFIX	"<!-- "
#define COMMENT_SUFFIX	" -->"

#if !defined(DEFAULT_DOCTYPE)
#define DEFAULT_DOCTYPE "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML//EN\">"
#endif

#if !defined(INDENT)
#define INDENT 4			/* can override at compile time */
#endif

#if !defined(MAXBUF)
#define MAXBUF 2048			/* can override at compile time */
#endif

#if !defined(MAXLISTNEST)
#define MAXLISTNEST	10		/* can override at compile time */
#endif

#if !defined(MAXWIDTH)
#define MAXWIDTH 72			/* can override at compile time */
#endif

#define NOOP	/* NO-OP */

#if defined(MSDOS)|| defined(_MSDOS)|| defined(__MSDOS__)|| defined(__TURBOC__)
#define IBM_PC
#endif

#if defined(IBM_PC)
#define OPTION_PREFIX(c)	( ((c) == '-') || ((c) == '/') )
/* On the IBM PC, we process files in binary mode, so that Ctl-Z does
not act as an end-of-file mark. */
#define RMODE "rb"
#define WMODE "wb"
#include <fcntl.h>			/* for O_BINARY in setmode() call */
#include <io.h>				/* for setmode() prototypes */

#if defined(_O_BINARY) && !defined(O_BINARY)
/* Microsoft C version 7.0 requires leading underscore on these symbols
in Standard C mode (command-line -Za switch).  Versions 5.0, 5.1, and
6.0 do not use the leading underscore.  */
#define O_BINARY _O_BINARY
#define setmode _setmode
#endif

#else  /* not IBM_PC */
#define OPTION_PREFIX(c)	((c) == '-')
#define RMODE "r"
#define WMODE "w"
#endif /* IBM_PC */

#include "stdc.h"			/* NB: must be included AFTER */
					/* all system header files, to */
					/* avoid inconsistent definitions */
					/* of const */

#if defined(__TURBOC__)

/* This code will compile fine under Turbo C 2.0, which always defines
fileno().  However, Turbo C 3.0 defines fileno() only in a non-Standard
C environment (-A command-line flag omitted).  Compilation fails under
both 2.0 and 3.0 if -A is omitted, because some header files get
included with const modifiers (flex checks for Turbo C and ensures that
const is used), and they later get included again with const defined to
an empty string (by stdc.h), resulting in function prototype declaration
conflicts.  Thus, we need to compile in the Standard C environment, and
then supply a definition of fileno() if one hasn't been provided. */

#ifndef fileno
#define fileno(f)       ((f)->fd)
#endif
#endif /* defined(__TURBOC__) */

/* Silicon Graphics IRIX 5.3 lex supplies a 200-character input
buffer, which is much too small, so increase it. */

#if defined(YYLMAX) && ((YYLMAX + 0) < 8192)
#undef YYLMAX
#define YYLMAX 8192
#endif

static char		big_buffer[2*MAXBUF]; /* size must be > MAXBUF */
static int		big_next_position = 0; /* index into big_buffer[] */
static int		body_count = 0;
static int		body_level = 0;
static int		do_banner = 1;
static int		doctype_count = 0;
static const char	*filename = "";
static int		g_argc;
static char		**g_argv;
static int		g_argk;
static int		g_errors;
static int		head_count = 0;
static int		head_level = 0;
static int		html_count = 0;
static int		html_level = 0;
static int		indent = INDENT;
static int		item_level[MAXLISTNEST+1]; /* index by 1..MAXLISTNEST */
static const char	*item_end_tag[MAXLISTNEST+1];
static int		level = 0;
static char		line_buffer[MAXBUF];
static int		line_number = 1;
static int		link_count = 0;
static int		list_level = 0;
static int		main_level = 0;
static int		max_width = MAXWIDTH;
static int		next_position = 0; /* index into line_buffer[] */
static int		paragraph_level = 0;

#if defined(DEBUG)
static char		prefix[4];	/* for DEBUGGING */
#endif

static int		title_count = 0;
static int		title_level = 0;

static void		adjust_level ARGS((int n_));
static void		author ARGS((void));
static void		begin_font ARGS((void));
static void		begin_list ARGS((void));
static void		begin_list_header ARGS((void));
static void		begin_list_item ARGS((void));
static void		begin_math ARGS((void));
static void		begin_pair ARGS((void));
static void		begin_paragraph ARGS((void));
static void		begin_verbatim ARGS((void));
static void		blank ARGS((void));
static void		body ARGS((void));
static void		comment ARGS((void));
static void		copyright ARGS((void));
static void		copy_verbatim ARGS((void));
static void		count_lines ARGS((void));
static void		delete_spaces ARGS((void));
static void		do_arg ARGS((void));
static void		do_begin_body ARGS((void));
static void		do_begin_head ARGS((void));
static void		do_begin_html ARGS((void));
static void		do_doctype ARGS((void));
static void		do_end_body ARGS((void));
static void		do_end_head ARGS((void));
static void		do_end_html ARGS((void));
static void		do_link ARGS((void));
static void		do_title ARGS((void));
static void		doctype ARGS((void));
static void		dputc ARGS((int c_));
static void		dputs ARGS((const char *s_, int length_));
static int		empty_paragraph ARGS((void));
static void		end_font ARGS((void));
static void		end_list ARGS((void));
static void		end_list_header ARGS((void));
static void		end_list_item ARGS((const char *tag_));
static void		end_math ARGS((void));
static void		end_pair ARGS((void));
static void		end_paragraph ARGS((void));
static void		end_verbatim ARGS((const char *tag_));
static void		error ARGS((const char *msg_));
static void		expect_level ARGS((int the_level_));
static void		font ARGS((void));
static void		head ARGS((void));
static void		hn ARGS((void));
static void		html ARGS((void));
static void		html_link ARGS((void));
static int		indentation_size ARGS((void));

#if defined(__cplusplus)
			/* Sun Solaris 2.x lex renames input() to yyinput() */
#define Input()		yyinput()
#else
#define Input()		input()
#endif /* defined(__cplusplus) */

#define Is_Begin_Tag(s)	((s)[1] != '/')

#if 0
#define Is_Comment_Tag(s) strncmp((s),"<!--",4) == 0) /* slow version */
#else
#define Is_Comment_Tag(s) ((s[0] == '<') && (s[1] == '!') && (s[2] == '-') && (s[3] == '-'))
					/* fast version */
#endif

#define Is_End_Tag(s)	(!Is_Begin_Tag(s))
#define Is_NonEmpty_Line()	(line_length() > indentation_size())

static int		last_char ARGS((int offset_));

#if defined(FLEX_SCANNER)
#define LEXEOF(c)	((c) == EOF)	/* flex end-of-file test */
#else
#define LEXEOF(c)	((c) == 0)	/* lex end-of-file test */
#endif

static void		line_break ARGS((void));
static void		line_end ARGS((void));
static int		line_length ARGS((void));
static void		list ARGS((void));
static void		list_header ARGS((void));
static void		list_item ARGS((void));
static void		math ARGS((void));

#define maybe_begin_body() ((body_count == 0) ? (do_begin_body(), 1) : 0)

static void		newline_after ARGS((void));
static void		newline_around ARGS((void));

#if 0
static void		newline_before ARGS((void));
#endif

static void		no_space_before ARGS((void));
static void		normalize_tag ARGS((char *tag_));
static void		other_tag ARGS((void));
static void		out_banner ARGS((void));
static void		out_begin_tag ARGS((const char *s_));
static void		out_blank ARGS((void));
static void		out_buffer ARGS((void));
static void		out_char ARGS((int c_));
static void		out_doctype ARGS((const char *s_));
static void		out_end_tag ARGS((const char *s_));
static void		out_escape ARGS((const char *s_));
static void		out_indentation ARGS((void));
static void		out_leading_blanks ARGS((int nblanks_));
static void		out_newline ARGS((void));
static void		out_string ARGS((const char *s_));
static void		out_tag_on_own_line ARGS((const char *tag_));

#define out_verbatim_char(c) (void)dputc(c)

static void		out_verbatim_string ARGS((const char *s_));
static void		out_word ARGS((void));
static void		out_yytext ARGS((void));
static void		pair ARGS((void));
static void		pair_nocheck ARGS((void));
static void		paragraph ARGS((void));
static int		paragraph_contains ARGS((const char *tag));
static void		punctuation ARGS((void));
static const char *	quote_check ARGS((const char *tag));
static int		reopen ARGS((void));
static void		short_tag ARGS((void));
static void		space_after ARGS((void));
static void		space_before ARGS((void));
static void		standalone ARGS((void));
static void		standalone_nocheck ARGS((void));
static void		title ARGS((void));
static void		trim_line ARGS((int n_));
static void		usage ARGS((void));
static void		verbatim ARGS((void));
static void		version ARGS((void));
static void		warning ARGS((const char *msg_));
static void		wrapup ARGS((void));

#if defined(yywrap)
#undef yywrap		/* this is a macro in some older flex versions */
#endif

#ifdef __cplusplus
extern "C"
#endif
int			yywrap ARGS((void));

%}

	/* These settings for lex are about double the required,  */
	/* during initial development.  They can be reduced when  */
	/* the code is stable.  */

%a	25000
%e	 5000
%k	 9000
%n	 2000
%o	40000
%p	10000

A	[Aa]
B	[Bb]
C	[Cc]
D	[Dd]
E	[Ee]
F	[Ff]
G	[Gg]
H	[Hh]
I	[Ii]
J	[Jj]
K	[Kk]
L	[Ll]
M	[Mm]
N	[Nn]
O	[Oo]
P	[Pp]
Q	[Qq]
R	[Rr]
S	[Ss]
T	[Tt]
U	[Uu]
V	[Vv]
W	[Ww]
X	[Xx]
Y	[Yy]
Z	[Zz]

BEGINPAIR	"<"[/]?
BEGINTAG	"<"

	/* SGML tags match "<NAME"{ENDTAG}, to allow for embedded */
	/* key="value" attributes.  Because this pattern matches */
	/* newline characters, every action that has ENDTAG in its */
	/* pattern MUST call count_lines() to update the line count. */
	/* Although AT&T lex has yylineno, POSIX lex does not, and */
	/* flex only provides it by special request.  It is unfortunate */
	/* that this requires tag tokens to be scanned more than once, */
	/* but fortunate that others do not require this overhead. */
ENDTAG		([^>A-Za-z0-9][^>]*)?">"

%%

[<]!--[^>]*--[>]				{ comment(); }


{BEGINPAIR}!{D}{O}{C}{T}{Y}{P}{E}{ENDTAG}	{ count_lines(); doctype(); }


{BEGINPAIR}{D}{I}{R}{ENDTAG}			|
{BEGINPAIR}{M}{E}{N}{U}{ENDTAG}			{ count_lines(); list(); } /* %HTML.Deprecated */


{BEGINPAIR}{D}{L}{ENDTAG}			|
{BEGINPAIR}{O}{L}{ENDTAG}			|
{BEGINPAIR}{U}{L}{ENDTAG}			{ count_lines(); list(); }


{BEGINPAIR}{L}{H}{ENDTAG}			{ count_lines(); list_header(); }


{BEGINPAIR}{D}{D}{ENDTAG}			|
{BEGINPAIR}{D}{T}{ENDTAG}			|
{BEGINPAIR}{L}{I}{ENDTAG}			{ count_lines(); list_item(); }


{BEGINPAIR}{A}{ENDTAG}				|
{BEGINPAIR}{A}{B}{B}{R}{E}{V}{ENDTAG}		|
{BEGINPAIR}{A}{B}{S}{T}{R}{A}{C}{T}{ENDTAG}	|
{BEGINPAIR}{A}{C}{R}{O}{N}{Y}{M}{ENDTAG}	|
{BEGINPAIR}{A}{D}{D}{E}{D}{ENDTAG}		|
{BEGINPAIR}{A}{D}{D}{R}{E}{S}{S}{ENDTAG}	|
{BEGINPAIR}{A}{R}{G}{ENDTAG}			|
{BEGINPAIR}{A}{R}{O}{W}{ENDTAG}			|
{BEGINPAIR}{A}{R}{R}{A}{Y}{ENDTAG}		|
{BEGINPAIR}{A}{U}{ENDTAG}			|
{BEGINPAIR}{B}{L}{O}{C}{K}{Q}{U}{O}{T}{E}{ENDTAG} |
{BEGINPAIR}{B}{Q}{ENDTAG}			|
{BEGINPAIR}{C}{A}{P}{T}{I}{O}{N}{ENDTAG}	|
{BEGINPAIR}{C}{I}{T}{E}{ENDTAG}			|
{BEGINPAIR}{C}{M}{D}{ENDTAG}			|
{BEGINPAIR}{D}{I}{V}1{ENDTAG}			|
{BEGINPAIR}{D}{I}{V}2{ENDTAG}			|
{BEGINPAIR}{D}{I}{V}3{ENDTAG}			|
{BEGINPAIR}{D}{I}{V}4{ENDTAG}			|
{BEGINPAIR}{D}{I}{V}5{ENDTAG}			|
{BEGINPAIR}{D}{I}{V}6{ENDTAG}			|
{BEGINPAIR}{F}{I}{G}{ENDTAG}			|
{BEGINPAIR}{F}{N}{ENDTAG}			|
{BEGINPAIR}{F}{O}{O}{T}{N}{O}{T}{E}{ENDTAG}	|
{BEGINPAIR}{F}{O}{R}{M}{ENDTAG}			|
{BEGINPAIR}{H}{I}{D}{E}{ENDTAG}			|
{BEGINPAIR}{L}{A}{N}{G}{ENDTAG}			|
{BEGINPAIR}{M}{A}{R}{G}{I}{N}{ENDTAG}		|
{BEGINPAIR}{M}{A}{T}{H}{ENDTAG}			|
{BEGINPAIR}{M}{E}{S}{S}{A}{G}{E}{ENDTAG}	|
{BEGINPAIR}{N}{O}{T}{E}{ENDTAG}			|
{BEGINPAIR}{O}{P}{T}{I}{O}{N}{ENDTAG}		|
{BEGINPAIR}{O}{V}{E}{R}{L}{A}{Y}{ENDTAG}	|
{BEGINPAIR}{P}{E}{R}{S}{O}{N}{ENDTAG}		|
{BEGINPAIR}{Q}{U}{O}{T}{E}{ENDTAG}		|
{BEGINPAIR}{R}{E}{M}{O}{V}{E}{D}{ENDTAG}	|
{BEGINPAIR}{S}{E}{L}{E}{C}{T}{ENDTAG}		|
{BEGINPAIR}{T}{A}{B}{L}{E}{ENDTAG}		|
{BEGINPAIR}{T}{E}{X}{T}{A}{R}{E}{A}{ENDTAG}	{ count_lines(); pair(); }


{BEGINPAIR}{B}{O}{D}{Y}{ENDTAG}			{ count_lines(); body(); }


{BEGINPAIR}{H}1{ENDTAG}				|
{BEGINPAIR}{H}2{ENDTAG}				|
{BEGINPAIR}{H}3{ENDTAG}				|
{BEGINPAIR}{H}4{ENDTAG}				|
{BEGINPAIR}{H}5{ENDTAG}				|
{BEGINPAIR}{H}6{ENDTAG}				{ count_lines(); hn(); }


{BEGINPAIR}{H}{E}{A}{D}{ENDTAG}			{ count_lines(); head(); }


{BEGINPAIR}{H}{T}{M}{L}{ENDTAG}			{ count_lines(); html(); }


{BEGINPAIR}{P}{ENDTAG}				{ count_lines(); paragraph(); }


{BEGINPAIR}{T}{I}{T}{L}{E}{ENDTAG}		{ count_lines(); title(); }


{BEGINPAIR}{B}{ENDTAG}				|
{BEGINPAIR}{B}{I}{G}{ENDTAG}			|
{BEGINPAIR}{C}{O}{D}{E}{ENDTAG}			|
{BEGINPAIR}{D}{F}{N}{ENDTAG}			|
{BEGINPAIR}{E}{M}{ENDTAG}			|
{BEGINPAIR}{I}{ENDTAG}				|
{BEGINPAIR}{K}{B}{D}{ENDTAG}			|
{BEGINPAIR}{Q}{ENDTAG}				|
{BEGINPAIR}{R}{E}{V}{ENDTAG}			|
{BEGINPAIR}{S}{ENDTAG}				|
{BEGINPAIR}{S}{A}{M}{P}{ENDTAG}			|
{BEGINPAIR}{S}{M}{A}{L}{L}{ENDTAG}		|
{BEGINPAIR}{S}{T}{R}{O}{N}{G}{ENDTAG}		|
{BEGINPAIR}{T}{T}{ENDTAG}			|
{BEGINPAIR}{U}{ENDTAG}				|
{BEGINPAIR}{V}{A}{R}{ENDTAG}			{ count_lines(); font(); }


{BEGINPAIR}{A}{B}{O}{V}{E}{ENDTAG}		|
{BEGINPAIR}{B}{E}{L}{O}{W}{ENDTAG}		|
{BEGINPAIR}{B}{O}{X}{ENDTAG}			|
{BEGINPAIR}{R}{O}{O}{T}{ENDTAG}			|
{BEGINPAIR}{S}{U}{B}{ENDTAG}			|
{BEGINPAIR}{S}{U}{P}{ENDTAG}			{ count_lines(); math(); }


{BEGINTAG}{B}{R}{ENDTAG}			{ count_lines(); line_break(); }


{BEGINTAG}{I}{T}{E}{M}{ENDTAG}			|
{BEGINTAG}{O}{V}{E}{R}{ENDTAG}			|
{BEGINTAG}{T}{A}{B}{ENDTAG}			|
{BEGINTAG}{T}{D}{ENDTAG}			|
{BEGINTAG}{T}{H}{ENDTAG}			|
{BEGINTAG}{T}{R}{ENDTAG}			{ count_lines(); short_tag(); }


{BEGINTAG}{C}{H}{A}{N}{G}{E}{D}{ENDTAG}		|
{BEGINTAG}{H}{R}{ENDTAG}			|
{BEGINTAG}{I}{M}{G}{ENDTAG}			|
{BEGINTAG}{I}{N}{P}{U}{T}{ENDTAG}		|
{BEGINTAG}{R}{E}{N}{D}{E}{R}{ENDTAG}		|
{BEGINTAG}{S}{T}{Y}{L}{E}{S}{ENDTAG}		{ count_lines(); standalone(); }


{BEGINTAG}{B}{A}{S}{E}{ENDTAG}			|
{BEGINTAG}{I}{S}{I}{N}{D}{E}{X}{ENDTAG}		|
{BEGINTAG}{M}{E}{T}{A}{ENDTAG}			|
{BEGINTAG}{S}{T}{Y}{L}{E}{ENDTAG}		{ count_lines(); standalone_nocheck(); }


{BEGINTAG}{N}{E}{X}{T}{I}{D}{ENDTAG}		{ count_lines(); standalone_nocheck(); } /* %HTML.Deprecated */


{BEGINTAG}{L}{I}{N}{K}{ENDTAG}			{ count_lines(); html_link(); }


{BEGINPAIR}{C}{D}{A}{T}{A}{ENDTAG}		{ count_lines(); verbatim(); } /* %HTML.Deprecated */


{BEGINPAIR}{L}{I}{S}{T}{I}{N}{G}{ENDTAG}	|
{BEGINPAIR}{P}{L}{A}{I}{N}{T}{E}{X}{T}{ENDTAG}	|
{BEGINPAIR}{P}{R}{E}{ENDTAG}			|
{BEGINPAIR}{X}{M}{P}{ENDTAG}			{ count_lines(); verbatim(); }


{BEGINPAIR}{B}{A}{R}{ENDTAG}			|
{BEGINPAIR}{D}{D}{O}{T}{ENDTAG}			|
{BEGINPAIR}{D}{E}{L}{ENDTAG}			|
{BEGINPAIR}{D}{O}{T}{ENDTAG}			|
{BEGINPAIR}{H}{A}{T}{ENDTAG}			|
{BEGINPAIR}{I}{N}{S}{ENDTAG}			|
{BEGINPAIR}{R}{O}{W}{ENDTAG}			|
{BEGINPAIR}{S}{Q}{R}{T}{ENDTAG}			|
{BEGINPAIR}{T}{I}{L}{D}{E}{ENDTAG}		|
{BEGINPAIR}{V}{E}{C}{ENDTAG}			{ count_lines(); math(); /* HTML 3.0 */ }


{BEGINPAIR}{B}{T}{ENDTAG}			|
{BEGINPAIR}{T}{ENDTAG}				{ count_lines(); font(); /* HTML 3.0 */ }


{BEGINPAIR}{C}{R}{E}{D}{I}{T}{ENDTAG}		{ count_lines(); pair(); /* HTML 3.0 */ }


{BEGINPAIR}[/]{ENDTAG}				|
{BEGINPAIR}[#A-Za-z][^>]*{ENDTAG}		{ count_lines(); other_tag(); }


[.,:;!?]					{ punctuation(); }


[&][#][0-9]+[;] |
[&][A-Za-z][A-Za-z0-9]*[;]			{ out_word(); }


[&]						{ out_escape("&amp;"); }


[<]						{ out_escape("&lt;"); }


[>]						{ out_escape("&gt;"); }


["]						{ out_escape("&quot;"); }


[^ <>"&\f\n\r\t\013]+		{	/* non-blank non-tag `words' */
					/* NB: lex on DEC ULTRIX, DEC OSF/1, */
					/* NeXT, and Sun SunOS 4.1.3 fails  */
					/* to handle \v (vertical tab) */
					/* correctly, so we replace it by */
					/* \013, sigh... */
					out_word();
				}


[ ]+				{ blank(); }


[\f\r\t\013]			{ out_blank(); }	/* other whitespace */


\n				{ line_end(); }


.				{ out_char((int)(yytext[0])); } /* other */


%%
static void
author(VOID_ARG)
{
    (void)fputs("Author:\n", stderr);
    (void)fputs("\tNelson H. F. Beebe\n", stderr);
    (void)fputs("\tCenter for Scientific Computing\n", stderr);
    (void)fputs("\tDepartment of Mathematics\n", stderr);
    (void)fputs("\tUniversity of Utah\n", stderr);
    (void)fputs("\tSalt Lake City, UT 84112\n", stderr);
    (void)fputs("\tUSA\n", stderr);
    (void)fputs("\tTel: +1 801 581 5254\n", stderr);
    (void)fputs("\tFAX: +1 801 581 4148\n", stderr);
    (void)fputs("\tEmail: <beebe@math.utah.edu>\n", stderr);
}


#if STDC
static void
adjust_level(int n)
#else
static void
adjust_level(n)
int n;
#endif
{
    level += n;
#if 0
    if (level < 0)
	error("negative indentation requested");
#else
    if (level < 0)
	level = 0;
#endif
}


static void
begin_font(VOID_ARG)
{
    maybe_begin_body();
    space_before();
}


static void
begin_list(VOID_ARG)
{
    maybe_begin_body();
    end_paragraph();		/* HTML lists cannot occur inside paragraphs */

    if (list_level < MAXLISTNEST)
    {
	++list_level;
	item_level[list_level] = 0;
	item_end_tag[list_level] = "";
    }
    else
	error("list nesting too deep: rebuild program with larger MAXLISTNEST");
    newline_around();
    adjust_level(1);
}


static void
begin_list_header(VOID_ARG)
{
    maybe_begin_body();
    end_paragraph();	/* HTML list headers cannot occur inside paragraphs */
    adjust_level(-1);
    newline_around();
}


static void
begin_list_item(VOID_ARG)
{
    maybe_begin_body();
    end_paragraph();		/* HTML list items cannot occur inside paragraphs */
    if (item_level[list_level] > 0)
	end_list_item(item_end_tag[list_level]); /* supply omitted end tag */
    adjust_level(-1);
    normalize_tag((char*)&yytext[0]);
    if (strncmp("<LI", yytext, 3) == 0)
	item_end_tag[list_level] = "</LI>";
    else if (strncmp("<DD", yytext, 3) == 0)
	item_end_tag[list_level] = "</DD>";
    else if (strncmp("<DT", yytext, 3) == 0)
	item_end_tag[list_level] = "</DT>";
    else
	error("unrecognized list item begin tag");
    newline_around();
    item_level[list_level]++;
}


static void
begin_math(VOID_ARG)
{
    maybe_begin_body();
    space_before();
}


static void
begin_pair(VOID_ARG)
{
    normalize_tag((char*)&yytext[0]);
    if (!paragraph_contains(yytext))
	end_paragraph();
    newline_around();
}


static void
begin_paragraph(VOID_ARG)
{
    maybe_begin_body();
    begin_pair();
    paragraph_level = level;
}


static void
begin_verbatim(VOID_ARG)
{
    maybe_begin_body();
    end_paragraph(); /* HTML verbatim environments cannot occur inside paragraphs */
    if ((level != main_level) && (list_level == 0))
	warning("verbatim environment at wrong level");
    trim_line(0);
    if (line_length() > 0)
	out_newline();
    trim_line(0);		/* want <VERBATIM> and </VERBATIM> on newline */
    out_verbatim_string((char*)&yytext[0]);
}


static void
blank(VOID_ARG)
{
    int c_last;
    int c_lastm1;
    int c_lastm2;

    delete_spaces();

    c_last = last_char(0);
    c_lastm1 = last_char(-1);
    c_lastm2 = last_char(-2);

    out_blank();		/* collapse multiple spaces into single one */

    switch (c_last)
    {
    case '.':
	if (isupper(c_lastm1))
	{
	    if (isspace(c_lastm2))
		break;			/* looks like [ ][A-Z][.] */
	    else
		out_blank();		/* assume sentence end */
	}
	else if ((c_lastm2 == '.') && isalpha(c_lastm1))
	    break;			/* looks like "e.g." or "i.e." */
	else
	    out_blank();		/* assume sentence end */
	break;

    case ':':
    case '!':
    case '?':
	out_blank();		/* two spaces after these punctuators */
	break;

    default:				/* no additional space */
	break;
    }
}


static void
body(VOID_ARG)
{
    if (Is_Begin_Tag(yytext))
    {
	do_begin_head();
	do_end_head();
	if (body_count > 0)
	    warning("illegal extra <BODY> tag");
	body_count++;
	body_level++;
    }
    else
	body_level--;

    pair_nocheck();
    expect_level(1);
    main_level = level;
}


static void
comment(VOID_ARG)
{
    newline_around();
    adjust_level(-1);
}


static void
copyright(VOID_ARG)
{
    (void)fprintf(stderr,"\t*****************************************\n");
    (void)fprintf(stderr,"\t* THIS PROGRAM IS IN THE PUBLIC DOMAIN. *\n");
    (void)fprintf(stderr,"\t*****************************************\n");
}


static void
copy_verbatim(VOID_ARG)
{
    /* Copy up to matching </ENDTAG>, or in the case of <PLAINTEXT>, */
    /* to end-of-file. */

    int c;
    char endtag[sizeof("</PLAINTEXT>")+1];
    char normalized_endtag[sizeof(endtag)];
    int k;
    char verbatim_name[sizeof("<PLAINTEXT>") + 1];
				/* space for longest verbatim tag name */

    (void)strncpy(verbatim_name, (char*)&yytext[0], sizeof(verbatim_name));
    verbatim_name[sizeof(verbatim_name)-1] = '\0';

    if (strcmp(verbatim_name,"<PLAINTEXT>") == 0)
	verbatim_name[1] = '\0'; /* prevent recognition of </PLAINTEXT> */

    out_buffer();			/* ensure output is up-to-date */

    while ((c = Input(), !LEXEOF(c)))
    {
	if (c == '\n')
	    line_number++;
	if (c == '<')
	{			/* try to collect tentative tag name */
	    endtag[0] = c;
	    for (k = 1; (k < (sizeof(endtag) - 1)) && (c != '>'); ++k)
	    {
		endtag[k] = c = Input();
		if (c == '\n')
		    line_number++;
	    }
	    endtag[k] = '\0';
	    if (c == '>')	/* found a complete tag */
	    {
		strcpy(normalized_endtag, endtag);
		normalize_tag((char*)&normalized_endtag[0]);
		if ((endtag[1] == '/') &&
		    (strcmp(&verbatim_name[1],&normalized_endtag[2]) == 0))
		{		/* found matching end tag */
		    normalize_tag((char*)&endtag[0]);
		    end_verbatim((const char*)&endtag[0]);
		    return;
		}
	    }
	    /* No matching tag, so just output our lookahead */
	    out_verbatim_string(endtag);
	}
	else
	    out_verbatim_char(c);
    }
}


static void
count_lines(VOID_ARG)
{
    register const char *p;

    for (p = (const char*)&yytext[0]; *p ; ++p)
    {
	if (*p == '\n')
	    line_number++;
    }
}


static void
delete_spaces(VOID_ARG)
{
    trim_line(indentation_size());
}


static void
do_begin_body(VOID_ARG)
{
    do_begin_head();
    do_end_head();
    if (body_count == 0)
    {
	out_begin_tag("<BODY>");
	body_count++;
	body_level++;
    }
}


static void
do_begin_head(VOID_ARG)
{
    do_begin_html();
    if (head_count == 0)
    {
	out_begin_tag("<HEAD>");
	head_count++;
	head_level++;
    }
}


static void
do_begin_html(VOID_ARG)
{
    do_doctype();
    if (html_count == 0)
    {
	out_begin_tag("<HTML>");
	html_count++;
	html_level++;
	main_level = level;
    }
}


static void
do_doctype(VOID_ARG)
{
    end_paragraph();
    if (doctype_count == 0)
	out_doctype(DEFAULT_DOCTYPE);
}


static void
do_end_body(VOID_ARG)
{
    if (body_level > 0)
    {
	end_paragraph();
	out_end_tag("</BODY>");
	body_level--;
    }
}


static void
do_end_head(VOID_ARG)
{
    if (head_level > 0)
    {
	end_paragraph();
	do_title();
	do_link();
	out_end_tag("</HEAD>");
	head_level--;
    }
}


static void
do_end_html(VOID_ARG)
{
    do_begin_body();
    do_end_body();
    if (html_level > 0)
    {
	out_end_tag("</HTML>");
	html_level--;
    }
}


static void
do_link(VOID_ARG)
{
    if (link_count == 0)
    {
	out_tag_on_own_line(
	    "<!-- Please supply a correct e-mail address here -->");
	out_tag_on_own_line(
	    "<LINK REV=\"made\" HREF=\"mailto:USERNAME@HOSTNAME\">");
	link_count++;
    }
}


static void
do_title(VOID_ARG)
{
    if (title_count == 0)
    {
	out_begin_tag("<TITLE>");
	out_tag_on_own_line("<!-- Please supply a descriptive title here -->");
	out_end_tag("</TITLE>");
	title_count++;
    }
}


static void
doctype(VOID_ARG)
{
    if (doctype_count > 0)
	warning("illegal extra <!DOCTYPE...> declaration");
    out_doctype(yytext);
}


#if STDC
static void
dputc(int c)
#else
static void
dputc(c)
int c;
#endif
{
    static char s[2] = " ";

    s[0] = (char)c;
    dputs(s,1);
}


#if STDC
static void
dputs(const char *s, int length)
#else
static void
dputs(s, length)
const char *s;
int length;
#endif
{
    if (s == (const char*)NULL)		/* special case: flush buffer */
    {
	(void)fputs(big_buffer, stdout);
	big_next_position = 0;
    }
    else
    {
	if ((big_next_position + length) >= sizeof(big_buffer))
	{				/* flush half of buffer */
	    char c;
	    register char *p;
	    register char *q;

	    q = &big_buffer[big_next_position/2];
	    c = *q;
	    *q = '\0';
	    (void)fputs(big_buffer, stdout);
	    *q = c;
	    for (p = big_buffer; *q; )	/* move last half to first half */
	        *p++ = *q++;		/* avoid strcpy() because of overlap */
	    big_next_position = (int)(p - big_buffer);
	}
	strcpy(&big_buffer[big_next_position], s);
	big_next_position += length;
    }
}


static int
empty_paragraph()
{			/* return 1 if big_buffer[] ends in empty paragraph */
    char *p;

    /* When </P> is encountered, scan backwards in the internal output buffer
       to see if we have an empty paragraph, and if so, delete back to the
       beginning of the line on which <P> occurs, and return 1. */

    if (big_next_position == 0)
	return (0);			/* empty buffer: cannot do test */

    for (p = &big_buffer[big_next_position-1];
	 (p >= (&big_buffer[0] + 2)) && isspace(*p); --p)
	/* NO-OP */;

    if ((p[-2] == '<') && (p[-1] == 'P') && (p[0] == '>'))
    {
	for (p = p - 3;
	     (p >= &big_buffer[0]) && (*p != '\n') && isspace(*p); --p)
	    /* NO-OP */;
	if (*p == '\n')
	{
	    big_next_position = (int)(p + 1 - big_buffer);
	    return (1);
	}
    }
    return (0);
}


static void
end_font(VOID_ARG)
{
    no_space_before();
    space_after();
}


static void
end_list(VOID_ARG)
{
    end_paragraph();
    if (list_level > 0)
    {
	if (item_level[list_level] > 0)
	    end_list_item(item_end_tag[list_level]);
	item_end_tag[list_level] = "";
	item_level[list_level] = 0;
	list_level--;
    }
    adjust_level(-1);
    newline_around();
}


static void
end_list_header(VOID_ARG)
{
    if (!paragraph_contains(yytext+1))
	end_paragraph();
    newline_around();
    adjust_level(1);
}


#if STDC
static void
end_list_item(const char *tag)
#else
static void
end_list_item(tag)
const char *tag;
#endif
{
    if (!paragraph_contains(yytext+1))
	end_paragraph();
    adjust_level(-1);
    out_tag_on_own_line(tag);
    item_level[list_level]--;
    adjust_level(1);
}


static void
end_math(VOID_ARG)
{
    space_after();
}


static void
end_pair(VOID_ARG)
{
    if (!paragraph_contains(yytext+1))
	end_paragraph();
    newline_around();
}


static void
end_paragraph(VOID_ARG)
{
    if (paragraph_level > 0)
    {
	if (level == paragraph_level)
	{
	    adjust_level(-1);
	    if (!empty_paragraph())
		out_tag_on_own_line("</P>");
	    paragraph_level = 0;
	}
	else
	    warning("unclosed previous paragraph");
    }
}


#if STDC
static void
end_verbatim(const char *tag)
#else
static void
end_verbatim(tag)
const char *tag;
#endif
{
    out_verbatim_string(tag);
    out_newline();
}


#if STDC
static void
error(const char *msg)
#else
static void
error(msg)
const char *msg;
#endif
{
    warning(msg);
    wrapup();
    exit (EXIT_FAILURE);
}


#if STDC
static void
expect_level(int the_level) /* the_level == expected level OUTSIDE <x>..</x> */
#else
static void
expect_level(the_level)
int the_level;
#endif
{
    char msg[sizeof("<averylongtag> at level 99 should be at level 99")+1];

    if ( (Is_End_Tag(yytext) && (level != the_level)) ||
	(Is_Begin_Tag(yytext) && ((level - 1) != the_level)) )
    {
	/* yytext[0] has been set to '\0' already: supply the original '<' */
	(void)sprintf(msg,
		      "<%.13s at level %d should be at level %d",
		      (char*)&yytext[1],
		      Is_End_Tag(yytext) ? level : (level - 1),
		      the_level);
	warning(msg);
    }
}


static void
font(VOID_ARG)
{
    if (Is_Begin_Tag(yytext))
	begin_font();
    else
	end_font();
}


static void
head(VOID_ARG)
{
    if (Is_Begin_Tag(yytext))
    {
	do_begin_html();
	if (head_count > 0)
	    warning("illegal extra <HEAD> tag");
	head_count++;
	head_level++;
    }
    else
    {
	do_title();
	if (Is_End_Tag(yytext))
	    do_link();
	head_level--;
    }
    pair_nocheck();
    expect_level(1);
    main_level = level;
}


static void
hn(VOID_ARG)
{
    maybe_begin_body();
    do_begin_body();
    if (Is_Begin_Tag(yytext))
	end_paragraph();
    pair();
    expect_level(main_level);
}


static void
html(VOID_ARG)
{
    if (Is_Begin_Tag(yytext))
    {
	if (html_count > 0)
	    warning("illegal extra <HTML> tag");
	html_count++;
	html_level++;
    }
    else
	html_level--;
    do_doctype();
    pair_nocheck();
    expect_level(0);
    main_level = level;
}


static void
html_link(VOID_ARG)
{
    link_count++;
    standalone_nocheck();
}


static int
indentation_size(VOID_ARG)
{
    int n;

    n = indent * level;

    /* Ensure that a reasonable amount of space is left on each line, even */
    /* if the indentation level gets large.   Without this cap, out_char() */
    /* can produce an infinite loop(): out_char() -> out_leading_blanks() */
    /* out_blank() -> out_char(). */

    if (n > (max_width - 16))
	n = (max_width - 16);
    if (n < 0)
	n = 0;
    return (n);
}


#if STDC
static int
last_char(int offset)
#else
static int
last_char(offset)
int offset;
#endif
{	/* return the character offset characters from the last written */
	/* or EOF if not in buffer */
    int k;

    k = next_position - 1 + offset;

    if ((k < 0) || (next_position <= k))
	return (EOF);
    else
	return ((int)(0xff & line_buffer[k]));
}


static void
line_break(VOID_ARG)
{
    maybe_begin_body();
    no_space_before();
    newline_after();
    adjust_level(-1);
}


static void
line_end(VOID_ARG)
{
    int c;
    int original_line_number;

    /* One or more consecutive empty lines mark a paragraph boundary, */
    /* whereas a single newline is just replaced by a space. */

    original_line_number = line_number;

    for (c = '\n'; !LEXEOF(c) && isspace(c); c = Input())
    {					/* gobble all following whitespace */
	if (c == '\n')
	    line_number++;
    }
    if (!LEXEOF(c))
	unput(c);			/* put back non-space lookahead */
    if (line_number > (original_line_number + 1))
    {					/* have paragraph break */
	trim_line(0);
	if (line_length() > 0)
	    out_newline();		/* two or more newlines mark a */
	out_newline();			/* paragraph boundary */
    }
    else				/* have simple line break */
    {
        delete_spaces();		/* discard unneeded space and */
	out_blank();			/* replace line break by one space */
    }
}


static int
line_length(VOID_ARG)
{
    return (next_position);
}


static void
list(VOID_ARG)
{
    if (Is_Begin_Tag(yytext))
	begin_list();
    else
	end_list();
}


static void
list_header(VOID_ARG)
{
    if (Is_Begin_Tag(yytext))
	begin_list_header();
    else
	end_list_header();
}


static void
list_item(VOID_ARG)
{
    if (Is_Begin_Tag(yytext))
	begin_list_item();
    else
	end_list_item((const char*)&yytext[0]);
}


static void
math(VOID_ARG)
{
    if (Is_Begin_Tag(yytext))
	begin_math();
    else
	end_math();
}


static void
newline_after(VOID_ARG)
{
    if (Is_End_Tag(yytext))		/* then have end tag: </TAG> */
	adjust_level(-1);
    out_yytext();
    if (Is_Begin_Tag(yytext))		/* then have begin tag: <TAG> */
	adjust_level(1);
    out_newline();
}


static void
newline_around(VOID_ARG)
{
    if (Is_End_Tag(yytext))
	adjust_level(-1);
    delete_spaces();
    if (Is_NonEmpty_Line())		/* then put this tag on new line */
	out_newline();
    out_yytext();
    if (Is_Begin_Tag(yytext))
	adjust_level(1);
    out_newline();
}


#if 0
static void
newline_before(VOID_ARG)
{
    if (Is_End_Tag(yytext))		/* then have end tag: </TAG> */
	adjust_level(-1);
    out_newline();
    out_yytext();
    if (Is_Begin_Tag(yytext))		/* then have begin tag: <TAG> */
	adjust_level(1);
}
#endif

static void
no_space_before(VOID_ARG)
{
    delete_spaces();
    out_yytext();
}


#if STDC
static void
normalize_tag(char *tag)
#else
static void
normalize_tag(tag)
char *tag;
#endif
{		/* convert tag name and key names in tag[] to uppercase, */
		/* replace line breaks by spaces, and collapse adjacent */
		/* spaces to one space, overwriting the old contents */
		/* with the (possibly shorter) new contents */
    char *equals;
    char *p;
    char *q;

    if (tag[0] != '<')
	return;

    if (Is_Comment_Tag(tag))		/* leave comment spacing untouched */
	return;

    p = (char*)&tag[1];
    if (*p == '/')
	++p;

    /* Convert tagname pattern "[A-Za-z]+" to uppercase */
    for ( ; *p && isalpha(*p) ; ++p)
    {
	if (islower(*p))
	    *p = toupper(*p);
    }

    /* Convert key name patterns "[A-Za-z]+=" to uppercase */
    while ((p != (char*)NULL) && ((equals = strchr(p, '=')) != (char*)NULL))
    {
	for (p = equals - 1; isalpha(*p); --p)
	{
	    if (islower(*p))
		*p = toupper(*p);
	}
	/* Try to scan past ="value" or ='value' in case the */
	/* value contains an equals.  However, some HTML/SGML markup */
	/* omits quotes around values, so in such a case, we just */
	/* scan for the next equals. */
	p = equals + 1;
	if ( (*p == '\"') || (*p == '\'') )
	    p = strchr(p+1, *p); /* advance to matching delimiter */
    }

    /* Collapse runs of whitespace to single blanks */
    for (p = q = (char*)&tag[0]; *q; ++p, ++q)
    {
	if (isspace(*q))
	{
	    while (*q && isspace(*q))	/* scan over run of spaces */
		++q;
	    if (*q == '\0')
		break;
	    *p++ = ' ';			/* and replace by a single blank */
	}
	*p = *q;
    }
    *p = '\0';				/* supply string terminator */
}


static void
other_tag(VOID_ARG)
{
    /* Unrecognized tags are normalized and output without indentation */
    /* change, but may force a paragraph end. */

    maybe_begin_body();
    normalize_tag((char*)&yytext[0]);
    if (Is_Begin_Tag(yytext))
    {
	if (!paragraph_contains(yytext))
	    end_paragraph();
    }
    else
    {
	if (!paragraph_contains(yytext+1))
	    end_paragraph();
    }
    out_yytext();
}


static void
out_banner(VOID_ARG)
{
    char	timestr[27];
    long	the_timeval;

    out_string(COMMENT_PREFIX);
    out_string("-*-html-*-");
    out_string(COMMENT_SUFFIX);
    out_newline();

    if (filename && *filename)
    {
	out_string(COMMENT_PREFIX);
	out_string(filename);
	out_string(COMMENT_SUFFIX);
	out_newline();
    }

    out_string(COMMENT_PREFIX);
    out_string("Prettyprinted by ");
    out_string(PROGRAM);
    out_string(" lex version ");
    out_string(VERSION);
    out_string(" ");
    out_string(DATE);
    out_string(COMMENT_SUFFIX);
    out_newline();

    the_timeval = time((time_t *)NULL);
    (void)strcpy(timestr,ctime((const time_t*)(&the_timeval)));
    timestr[24] = '\0';			/* kill \n from ctime() */
    out_string(COMMENT_PREFIX);
    out_string("on ");
    out_string(timestr);
    out_string(COMMENT_SUFFIX);
    out_newline();

#if defined(UNIX)
    {
	const char *login_name;
	const char *personal_name;
	struct passwd *pwd;

#if defined(HAVE_UNAME)
	struct utsname uts;
#else
	char hostname[256+1];
#endif

	login_name = (const char*)getlogin();
	personal_name = "";
	if (login_name == (const char*)NULL)
	    pwd = (struct passwd *)getpwuid(getuid());
	else
	    pwd = (struct passwd *)getpwnam(login_name);
	if (pwd != (struct passwd *)NULL)
	{
	    login_name = (const char*)pwd->pw_name;
	    personal_name = (const char*)pwd->pw_gecos;
	}
	else
	{
	    login_name = "";
	    personal_name = "";
	}

#if defined(HAVE_UNAME)
	if (uname(&uts) < 0)
	    (void)strcpy(uts.nodename,"");
#else
	if (gethostname(hostname, sizeof(hostname)) < 0)
	    (void)strcpy(hostname, "");
#endif

	out_string(COMMENT_PREFIX);
	out_string("for ");
	out_string(personal_name);
	out_string(" (");
	out_string(login_name);
	out_string("@");

#if defined(HAVE_UNAME)
	out_string(uts.nodename);
#else
	out_string(hostname);
#endif

	out_string(")");
	out_string(COMMENT_SUFFIX);
	out_newline();
    }
#endif /* defined(UNIX) */

    out_newline();			/* supply one blank line after banner */
}


#if STDC
static void
out_begin_tag(const char *s)
#else
static void
out_begin_tag(s)
const char *s;
#endif
{
    out_tag_on_own_line(s);
    adjust_level(1);
}


static void
out_blank(VOID_ARG)
{			/* output a blank, except at beginning of line */
    if (Is_NonEmpty_Line()) /* then not at start of line, */
	out_char((int)' ');
}


static void
out_buffer(VOID_ARG)
{
    if (next_position > 0)
    {
	line_buffer[next_position] = '\0';

#if defined(DEBUG)
	(void)fputs(prefix, stdout);	/* for DEBUGGING */
#endif

#if 0
	(void)fputs(line_buffer, stdout);
#else
	dputs(line_buffer,next_position);
#endif
	next_position = 0;
	line_buffer[next_position] = '\0'; /* for DEBUGGING only */
    }
}


#if STDC
static void
out_char(int c)
#else
static void
out_char(c)
int c;
#endif
{		/* ALL output goes through this routine, but it is buffered */
		/* in memory until a newline is written. */
    int wrapping;

    if ((next_position >= max_width) && (c == (int)' '))
    {					/* break long line here */
	c = '\n';
	wrapping = 1;
    }
    else
	wrapping = 0;

    if (c == '\n')		/* trim trailing space at end-of-line */
    {
	trim_line(0);

#if defined(IBM_PC)
	/* On the IBM PC, we handle input and output in binary mode, */
	/* and input \r (Ctl-M) gets mapped to blank by lex.  Thus, */
	/* on output, we need to map \n (Ctl-J) back to \r\n to match */
	/* DOS conventions.  This routine is never called from */
	/* copy_verbatim(), so there is no risk of getting \r\r\n */
	/* output from \r\n in verbatim mode. */
	line_buffer[next_position++] = '\r';
#endif

    }

    line_buffer[next_position++] = c;

    line_buffer[next_position] = '\0';	/* for DEBUGGING only */

    /* next_position must not advance beyond MAXBUF-3, because there must */
    /* be room to deposit three characters ('\r', '\n', '\0') on the next */
    /* call to this function */
    if ((c == '\n') || (next_position >= (MAXBUF - 3)))
    {					/* buffer full, or end of line */
	out_buffer();
	if (wrapping)
	    out_leading_blanks(indentation_size());
    }
}


#if STDC
static void
out_doctype(const char *s)
#else
static void
out_doctype(s)
const char *s;
#endif
{
    if (level > 0)
	warning("<!DOCTYPE...> not at level 0");
    doctype_count++;
    out_string(quote_check(s));
    out_newline();
    out_newline();	/* a following empty line improves visibility */
}


#if STDC
static void
out_end_tag(const char *s)
#else
static void
out_end_tag(s)
const char *s;
#endif
{
    adjust_level(-1);
    out_tag_on_own_line(s);
}


#if STDC
static void
out_escape(const char *s)
#else
static void
out_escape(s)
const char *s;
#endif
{
    maybe_begin_body();
    out_string(s);
}


static void
out_indentation(VOID_ARG)
{
    out_leading_blanks(indentation_size());
}


#if STDC
static void
out_leading_blanks(int nblanks)
#else
static void
out_leading_blanks(nblanks)
int nblanks;
#endif
{

#if defined(DEBUG)
    (void)sprintf(prefix, "%02d:", level);	/* for DEBUGGING */
#endif

    for ( ; nblanks > 0; --nblanks)
	out_char((int)' ');		/* NB: must NOT be out_blank() */
					/* because that function suppresses */
					/* blanks at start of line */
}


static void
out_newline(VOID_ARG)
{
    out_char((int)'\n');
}


#if STDC
static void
out_string(const char *s)
#else
static void
out_string(s)
const char *s;
#endif
{		/* output a string, line wrapping at spaces if necessary */
    const char *original_s;
    const char *p;

    if (line_length() < indentation_size()) /* first token on this line */
    {
	trim_line(0);
	out_indentation();
    }
    if (line_length() == indentation_size())
    {
	for (p = s; *p == ' '; ++p)	/* does s match /^[ ]+$/ ? */
	    NOOP;
	if (*p == '\0')		/* yes, it does */
	    return;		/* discard blank string at beginning of line */
    }
    for (original_s = s ; *s ; ++s)
    {
	if ((s == original_s) || (*s == ' '))
	{			/* then possible place to wrap the line */
	    p = strchr(s+1,' '); /* next wrap point is at next blank or */
	    if (p == (char*)NULL) /* else at end of string */
		p = strchr(s+1,'\0');
	    if (((line_length() + (int)(p - s)) > max_width) &&
		Is_NonEmpty_Line())
	    {			/* then line must be wrapped */
		out_newline();
		if (s > original_s)
		    adjust_level(1); /* indent wrapped lines one more level */
		out_indentation();
		if (s > original_s)
		    adjust_level(-1);
		if (*s == ' ')
		{
		    for (++s; *s && isspace(*s); ++s)
		        NOOP;		/* skip over repeated spaces */
		    s--;		/* and point back to last space */
		    continue;
		}
	    }
	}
	out_char(*s);
    }
}


#if STDC
static void
out_tag_on_own_line(const char *tag)
#else
static void
out_tag_on_own_line(tag)
const char *tag;
#endif
{
    delete_spaces();
    if (Is_NonEmpty_Line())		/* then put this tag on new line */
	out_newline();
    out_string(quote_check(tag));
    out_newline();
}


#if STDC
static void
out_verbatim_string(const char *s)
#else
static void
out_verbatim_string(s)
const char *s;
#endif
{
    for (; *s; ++s)
	out_verbatim_char((int)*s);
}


static void
out_word(VOID_ARG)
{
    if (title_level == 0)
	maybe_begin_body();
    out_yytext();
}


static void
out_yytext(VOID_ARG)
{
    if (yytext[0])
    {
	if (yytext[0] == '<')
	    normalize_tag((char*)&yytext[0]);

	if (Is_Comment_Tag(yytext))
	    out_verbatim_string((char*)&yytext[0]);
	else
	    out_string((yytext[0] == '<') ?
		       quote_check((const char*)&yytext[0]) :
		       (const char*)&yytext[0]);

	yytext[0] = '\0';		      /* prevent duplicate output */

	/* Having just reset yytext[] to an empty string, it appears
	that we should set yyleng = 0, but that is NOT allowed with
	flex.  In its quest for speed, flex by default requires that
	yytext[] must not be extended, and yyleng must never be changed
	by the user, unless the -l option (lex compatibility) has been
	used.  From (flex.info)Deficiencies: ``Use of `unput()'
	invalidates yytext and yyleng, unless the `%array' directive or
	the `-l' option has been used.''.  From (flex.info)OPtions:
	``Use of this option costs a considerable amount of
	performance...''.  In the current version of htmlpty.l, yyleng
	is never referenced, so it does not matter that we never update
	it.  However, the presence of this comment serves as a reminder
	for future revisions! */
    }
}


static void
pair(VOID_ARG)
{
    normalize_tag((char*)&yytext[0]);
    if (Is_Begin_Tag(yytext))
    {
	maybe_begin_body();
	begin_pair();
    }
    else
	end_pair();
}


static void
pair_nocheck(VOID_ARG)
{
    if (Is_Begin_Tag(yytext))
	begin_pair();
    else
	end_pair();
}


static void
paragraph(VOID_ARG)
{
    if (Is_Begin_Tag(yytext))
	begin_paragraph();
    else
	end_paragraph();
}


#if STDC
static int
paragraph_contains(const char *tag)
#else
static int
paragraph_contains(tag)
const char *tag;
#endif
{
    int k;
    static struct
    {
	const char *name;
	int length;
    } table[] =
    {

#define	ENTRY(s) { s, (sizeof(s) - 1) }

	ENTRY("#PCDATA"),
	ENTRY("A"),
	ENTRY("B"),
	ENTRY("BR"),
	ENTRY("CITE"),
	ENTRY("CODE"),
	ENTRY("DFN"),
	ENTRY("EM"),
	ENTRY("I"),
	ENTRY("IMG"),
	ENTRY("INPUT"),
	ENTRY("KBD"),
	ENTRY("SAMP"),
	ENTRY("STRONG"),
	ENTRY("TEXTAREA"),
	ENTRY("TT"),
	ENTRY("U"),
	ENTRY("VAR"),
	ENTRY("")
    };

    for (k = 0; table[k].length > 0; ++k)
    {
	if ((strncmp(tag+1,table[k].name,table[k].length) == 0) &&
	    ((tag[1+table[k].length] == '>') ||
	      isspace(tag[1+table[k].length])))
	    return (1);
    }
    return (0);
}


static void
punctuation(VOID_ARG)
{
    if (title_level == 0)
	maybe_begin_body();

    if ((last_char(0) == (int)' ') && (last_char(-1) == (int)'>'))
	delete_spaces();	/* discard space between tag and punctuation */
				/* e.g. "in <EM>html</EM>." */
    out_char(yytext[0]);
}


#if STDC
static const char *
quote_check(const char *tag)
#else
static const char *
quote_check(tag)
const char *tag;
#endif
{
    const char *close_angle;
    register const char *p;
    int quote_count;
    static char tempbuf[MAXBUF];

    if (tag[0] != '<')
	return (tag);

    if (Is_Comment_Tag(tag))		/* allow unbalanced quotes in comments */
	return (tag);			/* (e.g. in filehdr text) */

    for (close_angle = (const char*)NULL, p = tag, quote_count = 0; *p; ++p)
    {			/* count quotation marks, and remember tag end */
	if (*p == '"')
	    quote_count++;
	else if (*p == '>')
	    close_angle = p;
    }

    if ((quote_count % 2) == 0)
	return (tag);			/* okay: even number of quotes */
    else if ((close_angle != (const char*)NULL) &&
	     (strlen(tag) < (sizeof(tempbuf) - 1)))
    {					/* error: supply fixup */
	(void)strcpy(tempbuf, tag);
	tempbuf[(int)(close_angle - tag)] = '"';
	(void)strcpy(&tempbuf[(int)(close_angle - tag) + 1], close_angle);
	warning("unbalanced quotes in tag repaired");
	return ((const char*)&tempbuf[0]);
    }
    else				/* error: no space for fixup */
    {
	warning("unbalanced quotes in tag -- NOT repaired");
	return (tag);
    }
}


static void
short_tag(VOID_ARG)
{
    maybe_begin_body();
    out_yytext();
}


static void
space_after(VOID_ARG)
{
    out_yytext();
    out_blank();
}


static void
space_before(VOID_ARG)
{
    delete_spaces();
    out_blank();
    out_yytext();
}


static void
standalone(VOID_ARG)
{
    maybe_begin_body();
    standalone_nocheck();
}


static void
standalone_nocheck(VOID_ARG)
{
    normalize_tag((char*)&yytext[0]);
    if (!paragraph_contains(yytext))
	end_paragraph();
    newline_around();
    adjust_level(-1);	/* negate level adjustment from newline_around()  */
}


static void
title(VOID_ARG)
{
    if (Is_Begin_Tag(yytext))
    {
	do_begin_head();
	main_level = level;
	if (title_count > 0)
	    warning("illegal extra <TITLE> tag");
	title_count++;
	title_level++;
    }
    else
	title_level--;
    pair_nocheck();
    expect_level(main_level);
}


#if STDC
static void
trim_line(int n)
#else
static void
trim_line(n)
int n;
#endif
{
    for (; (next_position > 0) && (next_position > n) &&
	((line_buffer[next_position-1] == ' ') ||
	(line_buffer[next_position-1] == '\t'));
	 --next_position)
	NOOP;
}


static void
usage(VOID_ARG)
{
    static int first = 1;

    if (first)
    {
	(void)fputs(g_argv[0], stderr);
	(void)fputs(" [-?] [-a] [-c] [-f filename] [-h] [-i nnn] [-v] [-w width] [file(s)]\n",
		    stderr);
	(void)fputs("where the options are:\n", stderr);
	(void)fputs("\t[-?]          display usage summary\n", stderr);
	(void)fputs("\t[-a]          show author credit\n", stderr);
	(void)fputs("\t[-c]          show copyright information\n", stderr);
	(void)fputs("\t[-f filename] name output file in comment banner\n",
		    stderr);
	(void)fputs("\t[-h]          display usage summary\n", stderr);
	(void)fputs("\t[-i nnn]      set indentation to nnn spaces per level\n",
		    stderr);
	(void)fputs("\t[-n]          no comment banner\n", stderr);
	(void)fputs("\t[-v]          show version number\n", stderr);
	(void)fputs("\t[-w nnn]      set output line width to nnn\n", stderr);
    }
    first = 0;
}


static void
verbatim(VOID_ARG)
{
    normalize_tag((char*)&yytext[0]);
    if (Is_Begin_Tag(yytext))
    {
	begin_verbatim();
	copy_verbatim();
    }
    else
	end_verbatim((const char*)&yytext[0]);
}


static void
version(VOID_ARG)
{
    (void)fprintf(stderr, "HTML prettyprinter %s version %s\n",
		  PROGRAM, VERSION);
    (void)fprintf(stderr, "Compiled on %s %s by %s@%s\n", __DATE__, __TIME__,
		  USER, HOSTNAME);
}


#if STDC
static void
warning(const char *msg)
#else
static void
warning(msg)
const char *msg;
#endif
{
    (void)fprintf(stderr,"%s:%d:%s\n", filename, line_number, msg);
}


static void
wrapup(VOID_ARG)
{
    trim_line(0);
    if (line_length() > 0)		/* then force out last line */
	out_newline();
    do_end_html();
    dputs((const char*)NULL, 0);	/* flush internal buffer */
    if (level != 0)
	warning("non-zero <TAG> ... </TAG> nesting level at end-of-file");

}


#if STDC
int
main(int argc, char *argv[])
#else
int
main(argc, argv)
int argc;
char *argv[];
#endif
{

#if defined(IBM_PC)
    /* On IBM PC DOS, use binary mode for stdin and stdout, so that Ctl-Z */
    /* does not mean end-of-file on input, and so that Ctl-M Ctl-J is not */
    /* written at end-of-line on output, causing doubled Ctl-M characters. */
    (void)setmode(fileno(stdin), O_BINARY);
    (void)setmode(fileno(stdout), O_BINARY);
#endif

    do_banner = 1;
    g_errors = 0;
    g_argk = 0;
    g_argc = argc;
    g_argv = &argv[0];

    (void)reopen();

    if (do_banner)
	out_banner();

    while (yylex())
	NOOP;

    wrapup();				/* to force out last buffer */

    return ((g_errors > 0) ? EXIT_FAILURE : EXIT_SUCCESS);
}


static void
do_arg(VOID_ARG)
{
    int n;

    if (g_argv[g_argk][2] == '\0')
    {
	switch (g_argv[g_argk][1])
	{
	case 'a':
	case 'A':
	    author();
	    break;

	case 'c':
	case 'C':
	    copyright();
	    break;

	case 'f':
	case 'F':
	    filename = (const char*)g_argv[++g_argk];
	    if (filename == (const char *)NULL)
		filename = "";
	    break;

	case 'h':
	case 'H':
	case '?':
	    usage();
	    break;

	case 'i':
	case 'I':
	    if ((g_argv[++g_argk] != (char*)NULL) && isdigit(g_argv[g_argk][0]))
	    {
		n = (int)atoi(g_argv[g_argk]);
		if ((0 <= n) && (n <= 16))	/* require reasonable range */
		    indent = n;
	    }
	    else
	    {
		g_errors++;
		(void)fprintf(stderr,"Missing indentation value\n");
		usage();
	    }
	    break;

	case 'n':
	case 'N':
	    do_banner = 0;
	    break;

	case 'v':
	case 'V':
	    version();
	    break;

	case 'w':
	case 'W':
	    if ((g_argv[++g_argk] != (char*)NULL) && isdigit(g_argv[g_argk][0]))
	    {
		n = (int)atoi(g_argv[g_argk]);
		if ((16 <= n) && (n <= 256)) /* require reasonable range */
		    max_width = n;
	    }
	    else
	    {
		g_errors++;
		(void)fprintf(stderr,"Missing width value\n");
		usage();
	    }
	    break;

	default:
	    g_errors++;
	    (void)fprintf(stderr,
			  "Unrecognized option [%s] ignored\n",
			  g_argv[g_argk]);
	    usage();
	    break;
	}
    }
    else
    {
	g_errors++;
	(void)fprintf(stderr, "Unrecognized option [%s] ignored\n",
		      g_argv[g_argk]);
	usage();

    }
}


static int
reopen(VOID_ARG)
{
    if (g_argk < (g_argc - 1))
    {
	++g_argk;
	if (OPTION_PREFIX(g_argv[g_argk][0]))
	{
	    do_arg();
	    return (reopen());
	}
	else if (freopen(g_argv[g_argk], RMODE, stdin) != (FILE*)NULL)
	    filename = g_argv[g_argk];
	else
	    g_errors++;

	line_number = 1;
	return (0);			/* have file to process */
    }
    else
	return (1);			/* all done */
}


#ifdef __cplusplus
extern "C"
#endif
int
yywrap(VOID_ARG)
{
    return (reopen());
}
